/* ---------------------------------------------------------------------
     Small Music Information Retrieval Toolkit

    Copyright (c) 2008 Rebecca Fiebrink, Ge Wang and Perry R. Cook.  
    All rights reserved.
       http://smirk.cs.princeton.edu/
       http://chuck.cs.princeton.edu
       http://soundlab.cs.princeton.edu/

    Portions of SMIRK are based on the design of the Weka data mining
    toolkit, a GPL package.
       http://www.cs.waikato.ac.nz/ml/weka/
       Ian H. Witten and Eibe Frank (2005) "Data Mining: 
       Practical machine learning tools and techniques", 2nd Edition, 	
       Morgan Kaufmann, San Francisco, 2005.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A. 
--------------------------------------------------------------------- */

/**
 	AdaBoostM1: Multiclass AdaBoost
	Uses M1 algorithm to boost on decision stumps (for now)
*/
public class AdaBoostM1 extends Classifier
{    

	int m_t; //# of rounds (iterate from t=0 to (m_t-1) indices)
	float betas[0];
	DecisionStump dstumps[0];
	0 => int num_classes;
	0 => int max_class;
	0 => int min_class;
	0 => int num_iterations_performed; // will equal m_t if all training rounds succeed, or will be < m_t if training cut short due to high or 0 error
	1 => int show_progress; //make this a 0 if you don't want to watch training converge
	
	fun void set_num_rounds(int T) {
		T => m_t;
	}

	fun int get_num_rounds() {
		return m_t;
	}

	//overrides Classifier.train
    fun void train() {
		m_instances @=> Instances is;

		if (is.num_instances() == 0) {
			return;
		}
		
		m_instances.m_min_class => min_class;
		m_instances.m_max_class => max_class;
		(max_class - min_class) + 1 => num_classes;

		is.num_instances() => int num_train;
		int predictions[num_train][m_t];
		m_t => betas.size;
		float train_err[m_t];
	
		//initialize instance weights (Dt)
		for (0 => int i; i < num_train; i++) {
			1./num_train => is.m_instances[i].m_weight;
		}

		//Iterate through steps t=0 to T-1
		for (0 => int t; t < m_t; t++) {
			t => num_iterations_performed;
			<<< "******\n step ", t>>>;

			// Train weak learner on weighted instances
			DecisionStump ds;
			ds.set_instances(is);
			ds.train();
			dstumps << ds;
		
			0 => int num_correct_here;

			0. => float weighted_training_correct;
			for (0 => int i; i < num_train; i++) {
				is.m_instances[i] @=> Instance inst;
				ds.classify(inst) => predictions[i][t];
				if (predictions[i][t] == inst.m_class) {
					num_correct_here++;
				} //else {
				//	<<< "incorrect">>>;
				//}
				weighted_training_correct + inst.m_weight * (inst.m_class == predictions[i][t]) => weighted_training_correct;
			}

			1. - weighted_training_correct => float weighted_training_error;
			weighted_training_error => float epsilon;

			if (epsilon > .5 || epsilon == 0.) {
				<<< "Epsilon too high or 0; aborting on round ", t >>>;
				t => num_iterations_performed;
				if (t == 0) {
					1 => num_iterations_performed;
				}
				break;
			}	

			Math.log((1 - epsilon) / epsilon) => betas[num_iterations_performed];
			(1 - epsilon) / epsilon => float reweight;


			setWeights(reweight);

			//Maybe here it would be nice to print out progress in terms of training error
			if (show_progress) {
				0 => int num_correct;
				for (0 => int i; i < m_instances.num_instances(); i++) {
					classify(m_instances.m_instances[i]) => int prediction;
					if (prediction == m_instances.m_instances[i].m_class) {
						num_correct++;
					}
				}
				<<< num_correct, " of ", m_instances.num_instances(), " correct">>>;
			}
		}
	}

	fun void setWeights(float reweight) {
		float oldSumOfWeights;
		float newSumOfWeights;
		
		for (0 => int i; i < m_instances.num_instances(); i++) {
			m_instances.m_instances[i].m_weight + oldSumOfWeights => oldSumOfWeights;
		}
		
		for (0 => int i; i < m_instances.num_instances(); i++) {
			m_instances.m_instances[i] @=> Instance inst;
			dstumps[num_iterations_performed].classify(inst) => int prediction;
			if (prediction != inst.m_class) {
				inst.m_weight * reweight => inst.m_weight;
			}
		}

		//renormalize weights
		for (0 => int i; i < m_instances.num_instances(); i++) {
			m_instances.m_instances[i].m_weight + newSumOfWeights => newSumOfWeights;
		}

		for (0 => int i; i < m_instances.num_instances(); i++) {
			m_instances.m_instances[i] @=> Instance inst;
			inst.m_weight * oldSumOfWeights / newSumOfWeights => inst.m_weight;
		}

	}

	fun float[] distributionForInstance(Instance instance) {
		new float[num_classes] @=> float sums[];

		if (num_iterations_performed == 1) {
			return dstumps[0].distributionForInstance(instance);
		} else {
			for (0 => int i; i < num_iterations_performed; i++) {
				sums[norm_class_value(dstumps[i].classify(instance))] + betas[i] => sums[norm_class_value(dstumps[i].classify(instance))];
			}
			return logs2probs(sums);
		}
	}

	fun float[] logs2probs(float a[]) {
		0 => int maxIndex;
		-1 * Math.FLOAT_MAX => float maxValue;
		for (0 => int i; i < a.size(); i++) {
			if (a[i] > maxValue) {
				i => maxIndex;
				a[i] => maxValue;
			}
		}
		
		0 => float sum;
		new float[a.size()] @=> float result[];
		for (0 => int i; i < a.size(); i++) {
			Math.exp(a[i] - maxValue) => result[i];
			sum + result[i] => sum;
		}
		
		normalize(result, sum);
		return result;
		
	}

	fun void normalize(float floats[], float sum) {
		if (sum != 0) {
			for (0 => int i; i < floats.size(); i++) {
				floats[i] / sum => floats[i];
			}
		}
	}

    /* Override Classifier.classify */
    fun int classify(Instance tester) {
		if (num_iterations_performed == 0) {
			return 0;
		}

		distributionForInstance(tester) @=> float distribution[];

		0 => int maxIndex;
		distribution[0] => float maxVal;
		for (0 => int i; i < distribution.size(); i++) {
			if (distribution[i] > maxVal) {
				i => maxIndex;
				distribution[i] => maxVal;
			}
		}

		return real_class_value(maxIndex);

	} // end fun test
    

	fun int real_class_value(int i) {
		return (i + min_class);
	}

	//Normalize the class value to fall between 0 and num_classes-1, inclusive
    fun int norm_class_value(Instance i) {
		i.m_class => int real_class;
		return (real_class - min_class);
	}

	fun int norm_class_value(int i) {
		return (i - min_class);
	}

	//Override Classifier.clear
	fun void clear() {
		new Instances @=> m_instances;
		0 => num_iterations_performed;
		new float[0] @=> betas;
		new DecisionStump[0] @=> dstumps;
		0 => int num_classes;
		0 => int max_class;
		0 => int min_class;

	}
}