/* ---------------------------------------------------------------------
     Small Music Information Retrieval Toolkit

    Copyright (c) 2008 Rebecca Fiebrink, Ge Wang and Perry R. Cook.  
    All rights reserved.
       http://smirk.cs.princeton.edu/
       http://chuck.cs.princeton.edu
       http://soundlab.cs.princeton.edu/

    Portions of SMIRK are based on the design of the Weka data mining
    toolkit, a GPL package.
       http://www.cs.waikato.ac.nz/ml/weka/
       Ian H. Witten and Eibe Frank (2005) "Data Mining: 
       Practical machine learning tools and techniques", 2nd Edition, 	
       Morgan Kaufmann, San Francisco, 2005.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A. 
--------------------------------------------------------------------- */

/**
  Implements a MAUI GUI for AdaBoost.M1 classification
	Instructions: Edit the class names (first line below). 
	Change the features according to the classification problem.
	Press and hold the class name button to record instances of a class. 
	Set the number of boosting rounds using the slider and hit "Train" to train AdaBoost.
	Hold "Test" to run AdaBoost on the incoming audio. 
	Hit "Clear" to clear all recorded audio and start over.

	Only works with MiniAudicle on Mac.
*/

//Class names here
["A","B","C", "D", "E"] @=> string classes[];

// Patch for genre:
adc => FFT fft =^ Centroid c =^ FeatureCollector fc => blackhole;
fft =^ RMS r =^ fc;
fft =^ RollOff ro =^ fc;
fft =^ Flux flux =^ fc;

// set fft parameters
4096 => fft.size;
Windowing.hann(1024) => fft.window;

1 => int isFirst;

// compute srate
second / samp => float srate;

//Classifier and parameters
new AdaBoostM1 @=> AdaBoostM1 ab;
new Instances @=> Instances instances;
50 => int T; // # training rounds

//MAUI GUI elements
MAUI_View myWindow;
MAUI_Button recordb[classes.size()];
MAUI_LED which[classes.size()];
MAUI_Button trainb;
MAUI_Button testb;
MAUI_Slider slider;
MAUI_Button clear;
MAUI_Button exit;
build_gui();

//GUI logic
int i, j;
0 => int last_class;
1 => int running;
spork ~updateT();
while (running)	{
    0.050 :: second => now;

	//Check if any class is recording
	for (0=>j; j<classes.size(); j++)	{
	    if (recordb[j].state()==1)	{
			record(j);
			break;
		}
	}

	if (trainb.state() ==1 ) {
		train();
	} 
	
	if (testb.state() == 1) {
		test() => int tmp;
		if (tmp != last_class || isFirst) {
			which[last_class].unlight();
			which[tmp].light();
			0 => isFirst;
		}
		tmp => last_class;	
	} else {
		which[last_class].unlight();
		1 => isFirst;
	}
	
	if (clear.state() == 1) {
		ab.clear();
	}
	
	1 - exit.state() => running;
}

myWindow.destroy();
//End of program


// train on class c
function void record(int c) {
    
   	fc.upchuck() @=> UAnaBlob blob;
   	new Instance @=> Instance inst;
   	new float[blob.fvals().size()] @=> inst.m_fv;
        
   	// copy fvals to instance *by value*
   	// otherwise, they'll get overwritten by the next upchuck
   	for (0 => int j; j < inst.m_fv.size(); j++) {
       blob.fval(j) => inst.m_fv[j];
   	}
   	c => inst.m_class;
	instances.add(inst);        

    // advance time; no rush
    .05::second=>now;
} 

function void train() {
	instances.print();
	ab.set_instances(instances);
	ab.set_num_rounds(T);
	ab.train();	
	1 => isFirst;
}

fun void updateT() {
	while (true) {
		slider.onChange() => now;	
		slider.value()$int => T;
	}
}

function int test() {
    0 => int count1;
    0 => int count2;
    fc.upchuck() @=> UAnaBlob blob; 
    new Instance @=> Instance inst;
    new float[blob.fvals().size()] @=> inst.m_fv;        
    // copy fvals to instance *by value*
    for (0 => int j; j < inst.m_fv.size(); j++) {
        blob.fval(j) => inst.m_fv[j];
    }
    ab.classify(inst) => int c2;
   	return c2;        
}

fun void build_gui() {
	40 => float buttonHeight;
	30 => float textWidth;
	280 => float windowWidth;
	150 => float buttonWidth;
	myWindow.size(windowWidth,buttonHeight*classes.size() + 350);
	
	MAUI_Text t1;
	t1.name("Classes: Press to record instances");
	t1.position(0., 0.);
	t1.size(windowWidth, textWidth);
	myWindow.addElement(t1);

	MAUI_Text t;
	t.name("Classifier actions");
	t.position(0., buttonHeight*classes.size() + 50);
	t.size(windowWidth, textWidth);
	myWindow.addElement(t);

	trainb.name("Train");
	trainb.size(buttonWidth, 75.);
	trainb.position(0, buttonHeight*classes.size() + 70);
	testb.name("Test");
	testb.size(buttonWidth, 75.);
	testb.position(0, buttonHeight*classes.size() + 120);
	myWindow.addElement(testb);
	myWindow.addElement(trainb);

	clear.name("Clear");
	clear.position(0, buttonHeight*classes.size() + 170);
	clear.size(buttonWidth, 75.);
	myWindow.addElement(clear);


	(1, 100) => slider.range;
	T => slider.value;
	MAUI_Slider.integerFormat => slider.displayFormat;
	"Number of rounds" => slider.name;
	slider.position(0, buttonHeight*classes.size() + 220);
	myWindow.addElement(slider);

	exit.name("Exit");
	exit.position(0,buttonHeight*classes.size() + 270);
	exit.size(75, 75);
	myWindow.addElement(exit);
	
	int i,j;

	for (0=>j;j<classes.size();1+=>j)	{
    	recordb[j].name(classes[j]);
		recordb[j].size(buttonWidth,75.0);
		recordb[j].position(0.0,buttonHeight*j + 30);
		myWindow.addElement(recordb[j]);
		which[j].position(200,buttonHeight*j+30);
		myWindow.addElement(which[j]);
	}

	myWindow.name("Adaboost classifier");
	myWindow.display();
}


