/* ---------------------------------------------------------------------
     Small Music Information Retrieval Toolkit

    Copyright (c) 2008 Rebecca Fiebrink, Ge Wang and Perry R. Cook.  
    All rights reserved.
       http://smirk.cs.princeton.edu/
       http://chuck.cs.princeton.edu
       http://soundlab.cs.princeton.edu/

    Portions of SMIRK are based on the design of the Weka data mining
    toolkit, a GPL package.
       http://www.cs.waikato.ac.nz/ml/weka/
       Ian H. Witten and Eibe Frank (2005) "Data Mining: 
       Practical machine learning tools and techniques", 2nd Edition, 	
       Morgan Kaufmann, San Francisco, 2005.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A. 
--------------------------------------------------------------------- */

/**
   Multiple class AdaBoost classifier
	This one has a keyboard interface for 3 classes,
	but you can easily change it to a different number.

	Instructions for use:
		Edit the class names and feature extraction patch for your task
		Hold 1 to record training examples of class 1
		Hold 2 ... for class 2
		Hold 3 ... for class 3
		Press 4 to train the classifier
		Hold 5 to output the classifier's predictions
		Press 6 to clear the classifier
*/

//Add the class names here for output
string classes[3];
"led zeppelin" => classes[0];
"joni mitchell" => classes[1];
"stevie wonder" => classes[2];

// Feature extraction patch for genre:
adc => FFT fft =^ Centroid c =^ FeatureCollector fc => blackhole;
fft =^ RMS r =^ fc;
fft =^ RollOff ro =^ fc;
fft =^ Flux flux =^ fc;

// set fft parameters
4096 => fft.size;
Windowing.hann(1024) => fft.window;

//Initialize the classifier
new AdaBoostM1 @=> AdaBoostM1 ab;
50 => int T; // # training rounds

new Instances @=> Instances instances;

//Instantiate HID to get keyboard input
0 => int deviceNum;
HidIn hi;
HidMsg msg;

// open keyboard
if( !hi.openKeyboard( deviceNum ) ) me.exit();
<<< "keyboard '", hi.name(), "' ready" >>>;

// get keyboard input
0 => int stop;
0 => int going;
while (true) {
    //wait on event
    hi => now;
    //Get keyboard input
    while (hi.recv(msg)) {
        if (msg.isButtonDown()) {
            if (! going) {
                1 => going;
                0 => stop;
                
                if (msg.which == 30) {
                    //record class 1            
                    spork ~ record(1);
                }  else if (msg.which == 31) {
                    //record class 2
                    spork ~ record(2);
                } else if (msg.which == 32) {
                    //record class 3
                    spork ~ record(3);
                } else if (msg.which == 33) {
                    //train on all recorded
                    spork ~ train();
                } else if (msg.which == 34) {
                    //run in classify mode!
					spork ~ test();
                } else if (msg.which == 35) {
					<<< "*******CLEAR!!!******">>>;
					new AdaBoostM1 @=> ab;
					new Instances @=> instances;
				}
            }  
        } else {
            //key up: stop training/testing now
            1 => stop;
            0 => going;
        }
    }
} //repeat key input forever.

// train on class c
function void record(int c) {
    
    while (! stop) {        
        
        fc.upchuck() @=> UAnaBlob blob;
        new Instance @=> Instance inst;
        new float[blob.fvals().size()] @=> inst.m_fv;
        
        // copy fvals to instance *by value*
        // otherwise, they'll get overwritten by the next upchuck
        for (0 => int j; j < inst.m_fv.size(); j++) {
            blob.fval(j) => inst.m_fv[j];
        }
        c => inst.m_class;
	
        <<< "Record for class: ", c_name(c)>>>;
		instances.add(inst);        

        // advance time; no rush
        .05::second=>now;
    }

} 

function void train() {
	<<< "training on " , instances.num_instances() >>>;
	<<< "instances: ">>>;
	ab.set_instances(instances);
	ab.set_num_rounds(T);
	ab.train();	
}

function void test() {
    0 => int count1;
    0 => int count2;
    while (! stop) {
        fc.upchuck() @=> UAnaBlob blob; 
        new Instance @=> Instance inst;
        new float[blob.fvals().size()] @=> inst.m_fv;        
        // copy fvals to instance *by value*
        for (0 => int j; j < inst.m_fv.size(); j++) {
            blob.fval(j) => inst.m_fv[j];
        }
        
        ab.classify(inst) => int c2;
        if (c2 == 1) {
            count1 + 1=> count1;
        } else {
            count2 + 1=> count2;
        }
        <<<"class:", c_name(c2)>>>;

        //advance time, do this slower!
        .05::second => now;
    }
}

fun string c_name(int c) {
	if (c > 0 && c <= classes.size())
		return classes[c-1];
	else
		return "";
}
