/* ---------------------------------------------------------------------
     Small Music Information Retrieval Toolkit

    Copyright (c) 2008 Rebecca Fiebrink, Ge Wang and Perry R. Cook.  
    All rights reserved.
       http://smirk.cs.princeton.edu/
       http://chuck.cs.princeton.edu
       http://soundlab.cs.princeton.edu/

    Portions of SMIRK are based on the design of the Weka data mining
    toolkit, a GPL package.
       http://www.cs.waikato.ac.nz/ml/weka/
       Ian H. Witten and Eibe Frank (2005) "Data Mining: 
       Practical machine learning tools and techniques", 2nd Edition, 	
       Morgan Kaufmann, San Francisco, 2005.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A. 
--------------------------------------------------------------------- */

/**
	Decision stump classifier: A 1-level decision tree that uses
		weighted instances (weight affects entropy calculation to select splitting feature and split point)
	Algorithm from Weka's DecisionStump.java
*/

public class DecisionStump extends Classifier
{
    0 => int m_AttIndex;
    0. => float m_SplitPoint;

	//m_Distribution[0][] is the probability distribution for an instance where the split feature is below split point
	//m_Distribution[1][] is when feature is above split point
	float m_Distribution[2][1];
	0 => int num_classes;
	0 => int max_class;
	0 => int min_class;

	Math.log(2) => float log2;

	fun void clear() {
		new Instances @=> m_instances;
		new float[2][1] @=> m_Distribution;
		0 => num_classes;
		0 => max_class;
		0 => min_class;
		0 => m_AttIndex;
		0 => m_SplitPoint;
	}

    fun void train() {
		m_instances.m_min_class => min_class;
		m_instances.m_max_class => max_class;
		(max_class - min_class) + 1 => num_classes;

		Math.FLOAT_MAX => float bestVal;
		float currVal;
		-1 * Math.FLOAT_MAX => float bestPoint;
		-1 => int bestAtt;

		//The best distribution
		new float [2][num_classes] @=> float bestDist[][];
		1 => int is_first;

     	// Find best attribute to split
		for (0 => int i; i < m_instances.m_num_features; i++) {
           	//find best split for this feature
			new float[2][num_classes] @=> m_Distribution;
          		findSplit(i) => currVal;
          		if (is_first || (currVal < bestVal)) {
				currVal => bestVal;
				i => bestAtt;
				m_SplitPoint => bestPoint;

				for (0 => int j; j < m_Distribution[0].size(); j++) {
					m_Distribution[0][j] => bestDist[0][j];
					m_Distribution[1][j] => bestDist[1][j];
				}
			}
           	0 => is_first;
		}

		//Set attribute, split point, and distribution
		bestAtt => m_AttIndex;
		bestPoint => m_SplitPoint;
		bestDist @=> m_Distribution; //it's done here...

		for (0 => int i; i < m_Distribution.size(); i++) {
			sum(m_Distribution[i]) => float sumCounts;
			normalize(m_Distribution[i], sumCounts);
		}
    }

	fun float sum(float list[]) {
		0. => float sum;
		for (0 => int i; i < list.size(); i++) {
			sum + list[i] => sum;
		}
		return sum;
	}

    
	fun void normalize(float floats[], float sum) {
		if (sum != 0) {
			for (0 => int i; i < floats.size(); i++) {
				(floats[i] / sum) => floats[i];
			}
		}
	}

	fun float findSplit(int index) {
		Math.FLOAT_MAX => float bestVal;
		float currVal, currCutPoint;
		new float[num_classes] @=> float sum[];
		new float[2][num_classes] @=> float bestDist[][];

		//Compute counts for all values
		for (0 => int i; i < m_instances.num_instances(); i++) {
			m_instances.m_instances[i] @=> Instance inst;
			inst.m_weight + m_Distribution[1][norm_class_value(inst)] => m_Distribution[1][norm_class_value(inst)];
		}

		for (0 => int j; j < m_Distribution[0].size(); j++) {
			m_Distribution[1][j] @=> sum[j];
		}

		//save current distribution as best
		for (0 => int j; j < m_Distribution[0].size(); j++) {
			m_Distribution[0][j] => bestDist[0][j];
			m_Distribution[1][j] => bestDist[1][j];
		}

		//sort instances
		sort(index);

		//Make split counts for each possible split, and evaluate
		for (0 => int i; i < m_instances.num_instances() - 1; i++) {
			m_instances.m_instances[i] @=> Instance inst;
			m_instances.m_instances[i+1] @=> Instance instPlusOne;
			m_Distribution[0][norm_class_value(inst)] + inst.m_weight => m_Distribution[0][norm_class_value(inst)];
			m_Distribution[1][norm_class_value(inst)] - inst.m_weight => m_Distribution[1][norm_class_value(inst)];

			if (inst.m_fv[index] < instPlusOne.m_fv[index]) {
				(inst.m_fv[index] + instPlusOne.m_fv[index]) / 2.0 => currCutPoint;
				entropy(m_Distribution) => currVal;

				if (currVal < bestVal) {
					currCutPoint => m_SplitPoint;
					currVal => bestVal;
					for (0 => int j; j < m_Distribution[0].size(); j++) {
						m_Distribution[0][j] => bestDist[0][j];
						m_Distribution[1][j] => bestDist[1][j];
					}
				}
			}
		} // end for


		bestDist @=> m_Distribution;
		return bestVal;
	}
	
	fun int real_class_value(int i) {
		return (i + min_class);
	}

	//Normalize the class value to fall between 0 and num_classes-1, inclusive
    fun int norm_class_value(Instance i) {
		i.m_class => int real_class;
		return (real_class - min_class);
	}

	fun float entropy(float matrix[][]) {
		0 => float returnValue;
		float sumForRow;
		0 => float total;

		for (0 => int i; i < 2; i++) {
			0 => sumForRow;
			for (0 => int j; j < matrix[0].size(); j++) {
				returnValue + lnFunc(matrix[i][j]) => returnValue;
				matrix[i][j] + sumForRow => sumForRow;
			}
			returnValue - lnFunc(sumForRow) => returnValue;
			sumForRow + total => total;
		}
		if (closeToZero(total))
			return 0.0;
		
		return ((-1 * returnValue) / (total * log2));
	}

	fun int closeToZero(float num) {
		Math.pow(10,-6) => float small;
		if (0 - num < small && num < small)
			return 1;
		else
			return 0;	
	}

	fun float lnFunc(float num) {
		if (num < Math.pow(10,-6)) {
			return 0.0;
		} else {		
			return num * Math.log(num);
		}
	}

	fun float[] distributionForInstance(Instance instance) {
		return m_Distribution[whichSubset(instance)];
	}


	fun int whichSubset(Instance instance) {
		if (instance.m_fv[m_AttIndex] <= m_SplitPoint) {
			return 0;
		} else {
			return 1;
		}
	}

	//Overrides Classifiery classify
	fun int classify(Instance instance) {
		distributionForInstance(instance) @=> float distribution[];
		
		//the best class has the biggest distribution value
		0 => int maxIndex;
		distribution[0] => float maxValue;

		for (0 => int i; i < num_classes; i++) {
			if (distribution[i] > maxValue) {
				i => maxIndex;
				distribution[i] => maxValue;
			}
		}

		return real_class_value(maxIndex);
	}
    
	fun void sort(int index) {
		sort(index, m_instances);
	}

    
    //mergesort: this should be OK
    //would be nicer to have in-place sort to mitigate chuck disaster
    fun void sort (int f, Instances i) {
        mergesort(i.m_instances, 0, i.m_instances.size()-1, f); 
    }
    
    fun void merge(Instance list[], int p, int q, int r, int f) {
        0 => int i => int j;
        q - p + 1 => int n1;
        r - q => int n2;
        p => int k;
        Instance left[n1];
        Instance right[n2];
        for (; i < n1; i++) {
            list[p+i] @=> left[i];
        }
        for (; j < n2; j++) {
            list[q+j+1] @=> right[j];
        }
        0 => i => j;
        while (i < n1 && j < n2) {
            if (left[i].m_fv[f] < right[j].m_fv[f]) {
                left[i++] @=> list[k++];
            } else {
                right[j++] @=> list[k++];
            }
        }
        while (i < n1) {
            left[i++] @=> list[k++];
        }
        while (j < n2) {
            right[j++] @=> list[k++];
        }
    	}
    
    	fun void mergesort(Instance list[], int p, int r, int f) {
        if (p < r) {
            p + ((r-p)/2) => int q;
            mergesort(list, p, q, f);
            mergesort(list, q+1, r, f);
            merge(list, p, q, r, f);
        }
    	}

	fun void print_probs(Instances is) {
		<<< "f <= ", m_SplitPoint, ":">>>;
		for (0 => int i; i < m_Distribution[0].size(); i++) {
			<<< m_Distribution[0][i] >>>;
		}
		<<< "f > ", m_SplitPoint, ":">>>;
		for (0 => int i; i < m_Distribution[1].size(); i++) {
			<<< m_Distribution[1][i] >>>;
		}	
	} 


}



