/* ---------------------------------------------------------------------
     Small Music Information Retrieval Toolkit

    Copyright (c) 2008 Rebecca Fiebrink, Ge Wang and Perry R. Cook.  
    All rights reserved.
       http://smirk.cs.princeton.edu/
       http://chuck.cs.princeton.edu
       http://soundlab.cs.princeton.edu/

    Portions of SMIRK are based on the design of the Weka data mining
    toolkit, a GPL package.
       http://www.cs.waikato.ac.nz/ml/weka/
       Ian H. Witten and Eibe Frank (2005) "Data Mining: 
       Practical machine learning tools and techniques", 2nd Edition, 	
       Morgan Kaufmann, San Francisco, 2005.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A. 
--------------------------------------------------------------------- */

/**
	An example using kNN for 1-finger trackpad gesture identification
	Example gestures: circle, line, figure-8, clockwise versus counter-clockwise circle, etc.
	
	Instructions:
		To train on class 1, press and hold "1" while moving mouse; release "1" when done
		To train on class 2, press and hold "2" while moving mouse; release "2" when done
		To test, press and hold "3" while moving mouse; release when done
		You can iteratively train and test using 1-3.
		To clear memory and start over, press 4		

		IMPORTANT: each press/hold/release cycle will only train/test on ONE instance of a gesture
		You must train several times (at least twice) for each gesture
*/

new kNN @=> kNN k;
1 => k.m_num_neighbors;

//Instantiate HID to get keyboard input
0 => int deviceNum;
HidIn hi;
HidMsg msg;

// open keyboard
if( !hi.openKeyboard( deviceNum ) ) me.exit();
<<< "keyboard '", hi.name(), "' ready" >>>;

// instantiate a HidIn object for mouse
HidIn him;
// structure to hold HID messages
HidMsg msgm;

// open mouse 0, exit on fail
if( !him.openMouse( deviceNum ) ) me.exit();
// successful! print name of device
<<< "mouse '", him.name(), "' ready" >>>;

// get keyboard input
0 => int stop;
0 => int going;

Event stop_e;

while (true) {
    //wait on event
    hi => now;
    //Get keyboard input
    while (hi.recv(msg)) {
        if (msg.isButtonDown()) {
            if (! going) {
                1 => going;
                0 => stop;
                
                if (msg.which == 30) {
					<<< "train 1">>>;
                    //train class 1            
                    spork ~ go(1,1);
                }  else if (msg.which == 31) {
					<<< "train 2">>>;
                    //train class 2
                    spork ~ go(2,1);
                } else if (msg.which == 32) {
                    //test!
                    spork ~ go(0,0);
                } else if (msg.which == 33) {
                    //clear knn
                    <<< "*******CLEAR!!!*******">>>;
                    k.clear();
                }
            }  
        } else {
            //key up: stop training/testing now		
            1 => stop;
			stop_e.broadcast();
            0 => going;
        }
    }
} //repeat key input forever.

Event s1;

function void signal_stop() {
	stop_e => now;
	s1.broadcast();
}

function void signal_mouse() {
	him => now;
	s1.broadcast();	
}


// train/go on class c
function void go(int c, int is_train) {
	//wait until mousein or stop
	int mouse_x[1];
	int mouse_y[1];
	0 => mouse_x[0];
	0 => mouse_y[0];
	0 => int last_x;
	0 => int last_y;

	while (! stop) {

		//wait for either stop or mouse input
		spork ~ signal_stop();
		spork ~ signal_mouse();
		s1 => now;
	
    	if (! stop) {
			while( him.recv( msgm ) )
    		{
	        	if( msgm.isMouseMotion() )
    	    	{
        	    	// axis of motion
            		if( msgm.deltaX )
            		{
                			//<<< "mouse motion:", msgm.deltaX, "on x-axis" >>>;
							mouse_x << last_x + msgm.deltaX;
							mouse_y << last_y;
							msgm.deltaX + last_x => last_x;

            		}
            		else if( msgm.deltaY )
            		{
                		//<<< "mouse motion:", msgm.deltaY, "on y-axis" >>>;
						mouse_y << last_y + msgm.deltaY;
						mouse_x << last_x;
						msgm.deltaY + last_y => last_y;
            		}
        		} //else {
					//<<< "not motion:", msgm.type >>>;
				//}
   
	   		}
		}  // if not stop
	} //while not stop
	
	// sample 10 samples, evenly spread, from mouse input
	int sampledx[10];
	int sampledy[10];
	mouse_x.size() -1 => int max;

	//we want n points
	10 => int n;
	n $ float => float nf;
	for (0 => int i; i < n; i++) {
		//pick a point
		(((i $ float) * (1.0/nf) + .5 / (n $ float)) *  max) $ int => int next;
		mouse_x[next] => sampledx[i];
		mouse_y[next] => sampledy[i];
	}

	//To print out the sampled vector:
/*	<<< "sampled is:">>>;
	for (0 => int i; i < n; i++) {
		<<< sampledx[i], sampledy[i]>>>;
	} */

	//might want to normalize at this point, too?

	//train/test
    new Instance @=> Instance inst;
  	new float[n*2] @=> inst.m_fv;
        
    // copy fvals to instance *by value*
 	for (0 => int j; j < n; j++) {
    	sampledx[j] => inst.m_fv[2*j];
		sampledy[j] => inst.m_fv[2*j + 1]; 
    }
	if (is_train) {
	    c => inst.m_class;
         <<< "TRAIN: ", c>>>;
	 	k.add_instance(inst);	
	} else {
		<<< "TEST:">>>;
		<<< k.classify(inst) >>>;
	}
} 
