/* ---------------------------------------------------------------------
     Small Music Information Retrieval Toolkit

    Copyright (c) 2008 Rebecca Fiebrink, Ge Wang and Perry R. Cook.  
    All rights reserved.
       http://smirk.cs.princeton.edu/
       http://chuck.cs.princeton.edu
       http://soundlab.cs.princeton.edu/

    Portions of SMIRK are based on the design of the Weka data mining
    toolkit, a GPL package.
       http://www.cs.waikato.ac.nz/ml/weka/
       Ian H. Witten and Eibe Frank (2005) "Data Mining: 
       Practical machine learning tools and techniques", 2nd Edition, 	
       Morgan Kaufmann, San Francisco, 2005.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A. 
--------------------------------------------------------------------- */

/**
	kNN: k-nearest-neighbor classifier
*/
public class kNN extends Classifier
{
	//The number of neighbors (k)
    10 => int m_num_neighbors;
    fun int num_neighbors() { return m_num_neighbors; }
	0 => int num_classes;
	0 => int max_class;
	0 => int min_class;

	//Classify an unseen instance
	//Override Classifier classify
	fun int classify(Instance tester) {
        // first make array of (ID, distance) pairs
        new ID_Dist[m_instances.m_instances.size()] @=> ID_Dist dist_list[];

		m_instances.m_min_class => min_class;
		m_instances.m_max_class => max_class;
		(max_class - min_class) + 1 => num_classes;

		//Add an (ID, distance) entry for each instance in the training set
        for (0 => int i; i < m_instances.m_instances.size(); i++) {
            m_instances.m_instances[i] @=> Instance next;
            ID_Dist next_id_dist;
            i => next_id_dist.id;
            distance(tester, next) => next_id_dist.dist;
            next_id_dist @=> dist_list[i];
        }     

        //sort (ID, distance) list in increasing order of distance
        sort(dist_list);
        
        //take top num_neighbors (or fewer if few instances), compute counts of classes.
        new int[num_classes] @=> int classes[];
        0 => int max_count;
        0 => int max_class;
        Math.min(m_num_neighbors, m_instances.m_instances.size()) $ int => int num_to_use;
        for (0 => int i; i < num_to_use; i++) {
            dist_list[i].id => int id;
            m_instances.m_instances[id].m_class => int this_class;
			norm_class_value(this_class) => int this_norm_class;
            if (this_norm_class > classes.size() || this_norm_class < 0) {
				<<< "Error: Illegal class value ", this_class >>>;
				return -1;
			}
            classes[this_norm_class]++;
            if (classes[this_norm_class] > max_count) {
                this_norm_class => max_class;
                classes[this_norm_class] => max_count;
            } 
        }
        
        //here's the class with the majority of the k nearest neighbors
        return real_class_value(max_class);
    }

    
	//Helper function, computes squared Euclidean distance between two feature vectors
    //requires that a and b have the same number of features
    fun float distance(Instance a, Instance b) {
        0 => float dist;
		if (a.m_fv.size() != b.m_fv.size()) {
			<<< "Error: Distance requires that both Instance objects have same number of features" >>>;
			return -1.;
		}
		
        for (0 => int i; i < a.m_fv.size(); i++) {
            Math.pow(Std.fabs(a.m_fv[i] - b.m_fv[i]),2) +=> dist;
        }
        //leave out square root (unnecessary)
        return dist;
    }
    
	//Helper function: Perform mergesort to sort all distances
    fun void sort (ID_Dist list[]) {
        mergesort(list, 0, list.size()-1);
    }
    
	//Perform the merge step in mergesort
    fun void merge(ID_Dist list[], int p, int q, int r) {
        0 => int i => int j;
        q - p + 1 => int n1;
        r - q => int n2;
        p => int k;
        ID_Dist left[n1];
        ID_Dist right[n2];
        for (; i < n1; i++) {
            list[p+i] @=> left[i];
        }
        for (; j < n2; j++) {
            list[q+j+1] @=> right[j];
        }
        0 => i => j;
        while (i < n1 && j < n2) {
            if (left[i].dist < right[j].dist) {
                left[i++] @=> list[k++];
            } else {
                right[j++] @=> list[k++];
            }
        }
        while (i < n1) {
            left[i++] @=> list[k++];
        }
        while (j < n2) {
            right[j++] @=> list[k++];
        }
    }
    
	//Mergesort a list, from index p to index r
    fun void mergesort(ID_Dist list[], int p, int r) {
        if (p < r) {
            p + ((r-p)/2) => int q;
            mergesort(list, p, q);
            mergesort(list, q+1, r);
            merge(list, p, q, r);
        }
    } 
	
	fun int real_class_value(int i) {
		return (i + min_class);
	}

	//Normalize the class value to fall between 0 and num_classes-1, inclusive
    fun int norm_class_value(int i) {
		return (i - min_class);
	}

  
}